@@ -15,11 +15,15 @@ autoconf --version
 echo ""
 automake --version
 echo ""
-if test "`uname -s`" = Darwin; then
-  glibtool --version
+
+# Typical MacOS X installations rename 'libtoolize' to 'glibtoolize', so
+# adjust to that.
+if test "`uname -s`" = "Darwin"; then
+  glibtoolize --version
 else
-  libtool --version
+  libtoolize --version
 fi
+
 echo ""
 ${M4:-m4} --version
 echo ""
