@@ -1,5 +1,5 @@
-/* Copyright (C) 1995, 1996, 1997, 1999, 2000, 2001, 2003, 2004, 2006,
- *   2007, 2008, 2009, 2010, 2011, 2012, 2014 Free Software Foundation, Inc.
+/* Copyright (C) 1995-1997, 1999-2001, 2003, 2004, 2006-2012, 2014, 2015
+ *   Free Software Foundation, Inc.
  * 
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -25,13 +25,13 @@
 #endif
 
 #include <stdio.h>
-#include <ctype.h>
 #include <string.h>
 #include <unistd.h>
 #include <unicase.h>
 #include <unictype.h>
 #include <c-strcase.h>
 #include <c-ctype.h>
+#include <alloca.h>
 
 #include "libguile/_scm.h"
 #include "libguile/bytevectors.h"
@@ -45,8 +45,8 @@
 #include "libguile/hashtab.h"
 #include "libguile/hash.h"
 #include "libguile/ports.h"
+#include "libguile/ports-internal.h"
 #include "libguile/fports.h"
-#include "libguile/root.h"
 #include "libguile/strings.h"
 #include "libguile/strports.h"
 #include "libguile/vectors.h"
@@ -64,6 +64,7 @@ SCM_GLOBAL_SYMBOL (scm_sym_dot, ".");
 SCM_SYMBOL (scm_keyword_prefix, "prefix");
 SCM_SYMBOL (scm_keyword_postfix, "postfix");
 SCM_SYMBOL (sym_nil, "nil");
+SCM_SYMBOL (sym_ISO_8859_1, "ISO-8859-1");
 
 /* SRFI-105 curly infix expression support */
 SCM_SYMBOL (sym_nfx, "$nfx$");
@@ -148,8 +149,8 @@ scm_i_input_error (char const *function,
   scm_simple_format (string_port,
 		     scm_from_locale_string ("~A:~S:~S: ~A"),
 		     scm_list_4 (fn,
-				 scm_from_long (SCM_LINUM (port) + 1),
-				 scm_from_int (SCM_COL (port) + 1),
+				 scm_sum (scm_port_line (port), SCM_INUM1),
+				 scm_sum (scm_port_column (port), SCM_INUM1),
 				 scm_from_locale_string (message)));
     
   string = scm_get_output_string (string_port);
@@ -411,6 +412,11 @@ static SCM
 maybe_annotate_source (SCM x, SCM port, scm_t_read_opts *opts,
                        long line, int column)
 {
+  /* This condition can be caused by a user calling
+     set-port-column!.  */
+  if (line < 0 || column < 0)
+    return x;
+
   if (opts->record_positions_p)
     scm_i_set_source_properties_x (x, line, column, SCM_FILENAME (port));
   return x;
@@ -428,8 +434,8 @@ scm_read_sexp (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
                                    : ')'));
 
   /* Need to capture line and column numbers here. */
-  long line = SCM_LINUM (port);
-  int column = SCM_COL (port) - 1;
+  long line = scm_to_long (scm_port_line (port));
+  int column = scm_to_int (scm_port_column (port)) - 1;
 
   c = flush_ws (port, opts, FUNC_NAME);
   if (terminating_char == c)
@@ -606,8 +612,8 @@ scm_read_string_like_syntax (int chr, SCM port, scm_t_read_opts *opts)
   scm_t_wchar c, c_str[READER_STRING_BUFFER_SIZE];
 
   /* Need to capture line and column numbers here. */
-  long line = SCM_LINUM (port);
-  int column = SCM_COL (port) - 1;
+  long line = scm_to_long (scm_port_line (port));
+  int column = scm_to_int (scm_port_column (port)) - 1;
 
   while (chr != (c = scm_getc (port)))
     {
@@ -635,6 +641,9 @@ scm_read_string_like_syntax (int chr, SCM port, scm_t_read_opts *opts)
               goto str_eof;
             case '|':
             case '\\':
+            case '(':  /* Accept "\(" for use at the beginning of lines
+			  in multiline strings to avoid confusing emacs
+			  lisp modes.  */
               break;
             case '\n':
               if (opts->hungry_eol_escapes_p)
@@ -728,17 +737,16 @@ scm_read_number (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
   SCM result, str = SCM_EOL;
   char local_buffer[READER_BUFFER_SIZE], *buffer;
   size_t bytes_read;
-  scm_t_port *pt = SCM_PTAB_ENTRY (port);
 
   /* Need to capture line and column numbers here. */
-  long line = SCM_LINUM (port);
-  int column = SCM_COL (port) - 1;
+  long line = scm_to_long (scm_port_line (port));
+  int column = scm_to_int (scm_port_column (port)) - 1;
 
   scm_ungetc (chr, port);
   buffer = read_complete_token (port, opts, local_buffer, sizeof local_buffer,
 				&bytes_read);
 
-  str = scm_from_stringn (buffer, bytes_read, pt->encoding, pt->ilseq_handler);
+  str = scm_from_port_stringn (buffer, bytes_read, port);
 
   result = scm_string_to_number (str, SCM_UNDEFINED);
   if (scm_is_false (result))
@@ -751,7 +759,9 @@ scm_read_number (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
   else if (SCM_NIMP (result))
     result = maybe_annotate_source (result, port, opts, line, column);
 
-  SCM_COL (port) += scm_i_string_length (str);
+  scm_set_port_column_x (port,
+                         scm_sum (scm_port_column (port),
+                                  scm_string_length (str)));
   return result;
 }
 
@@ -763,7 +773,6 @@ scm_read_mixed_case_symbol (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
   size_t bytes_read;
   int postfix = (opts->keyword_style == KEYWORD_STYLE_POSTFIX);
   char local_buffer[READER_BUFFER_SIZE], *buffer;
-  scm_t_port *pt = SCM_PTAB_ENTRY (port);
   SCM str;
 
   scm_ungetc (chr, port);
@@ -774,8 +783,7 @@ scm_read_mixed_case_symbol (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
 
   if (postfix && ends_with_colon && (bytes_read > 1))
     {
-      str = scm_from_stringn (buffer, bytes_read - 1,
-			      pt->encoding, pt->ilseq_handler);
+      str = scm_from_port_stringn (buffer, bytes_read - 1, port);
 
       if (opts->case_insensitive_p)
         str = scm_string_downcase_x (str);
@@ -783,15 +791,16 @@ scm_read_mixed_case_symbol (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
     }
   else
     {
-      str = scm_from_stringn (buffer, bytes_read,
-			      pt->encoding, pt->ilseq_handler);
+      str = scm_from_port_stringn (buffer, bytes_read, port);
 
       if (opts->case_insensitive_p)
         str = scm_string_downcase_x (str);
       result = scm_string_to_symbol (str);
     }
 
-  SCM_COL (port) += scm_i_string_length (str);
+  scm_set_port_column_x (port,
+                         scm_sum (scm_port_column (port),
+                                  scm_string_length (str)));
   return result;
 }
 
@@ -804,7 +813,6 @@ scm_read_number_and_radix (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
   char local_buffer[READER_BUFFER_SIZE], *buffer;
   unsigned int radix;
   SCM str;
-  scm_t_port *pt;
 
   switch (chr)
     {
@@ -837,12 +845,13 @@ scm_read_number_and_radix (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
   buffer = read_complete_token (port, opts, local_buffer, sizeof local_buffer,
 				&read);
 
-  pt = SCM_PTAB_ENTRY (port);
-  str = scm_from_stringn (buffer, read, pt->encoding, pt->ilseq_handler);
+  str = scm_from_port_stringn (buffer, read, port);
 
   result = scm_string_to_number (str, scm_from_uint (radix));
 
-  SCM_COL (port) += scm_i_string_length (str);
+  scm_set_port_column_x (port,
+                         scm_sum (scm_port_column (port),
+                                  scm_string_length (str)));
 
   if (scm_is_true (result))
     return result;
@@ -857,8 +866,8 @@ static SCM
 scm_read_quote (int chr, SCM port, scm_t_read_opts *opts)
 {
   SCM p;
-  long line = SCM_LINUM (port);
-  int column = SCM_COL (port) - 1;
+  long line = scm_to_long (scm_port_line (port));
+  int column = scm_to_int (scm_port_column (port)) - 1;
 
   switch (chr)
     {
@@ -904,8 +913,8 @@ static SCM
 scm_read_syntax (int chr, SCM port, scm_t_read_opts *opts)
 {
   SCM p;
-  long line = SCM_LINUM (port);
-  int column = SCM_COL (port) - 1;
+  long line = scm_to_long (scm_port_line (port));
+  int column = scm_to_int (scm_port_column (port)) - 1;
 
   switch (chr)
     {
@@ -1056,23 +1065,26 @@ scm_read_character (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
       return (SCM_MAKE_CHAR (chr));
     }
 
-  pt = SCM_PTAB_ENTRY (port);
+  pt = SCM_PORT (port);
 
   /* Simple ASCII characters can be processed immediately.  Also, simple
      ISO-8859-1 characters can be processed immediately if the encoding for this
      port is ISO-8859-1.  */
-  if (bytes_read == 1 && ((unsigned char) buffer[0] <= 127 || pt->encoding == NULL))
+  if (bytes_read == 1 &&
+      ((unsigned char) buffer[0] <= 127
+       || scm_is_eq (pt->encoding, sym_ISO_8859_1)))
     {
-      SCM_COL (port) += 1;
+      scm_set_port_column_x (port, scm_sum (scm_port_column (port), SCM_INUM1));
       return SCM_MAKE_CHAR (buffer[0]);
     }
 
   /* Otherwise, convert the buffer into a proper scheme string for
      processing.  */
-  charname = scm_from_stringn (buffer, bytes_read, pt->encoding,
-			       pt->ilseq_handler);
+  charname = scm_from_port_stringn (buffer, bytes_read, port);
   charname_len = scm_i_string_length (charname);
-  SCM_COL (port) += charname_len;
+  scm_set_port_column_x (port,
+                         scm_sum (scm_port_column (port),
+                                  scm_from_size_t (charname_len)));
   cp = scm_i_string_ref (charname, 0);
   if (charname_len == 1)
     return SCM_MAKE_CHAR (cp);
@@ -1418,6 +1430,12 @@ static void set_port_square_brackets_p (SCM port, scm_t_read_opts *opts,
                                         int value);
 static void set_port_curly_infix_p (SCM port, scm_t_read_opts *opts,
                                     int value);
+static void set_port_r6rs_hex_escapes_p (SCM port, scm_t_read_opts *opts,
+                                         int value);
+static void set_port_hungry_eol_escapes_p (SCM port, scm_t_read_opts *opts,
+                                           int value);
+static void set_port_keyword_style (SCM port, scm_t_read_opts *opts,
+                                    enum t_keyword_style value);
 
 static SCM
 scm_read_shebang (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
@@ -1439,7 +1457,13 @@ scm_read_shebang (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
           scm_ungetc (c, port);
           name[i] = '\0';
           if (0 == strcmp ("r6rs", name))
-            ;  /* Silently ignore */
+            {
+              set_port_case_insensitive_p (port, opts, 0);
+              set_port_r6rs_hex_escapes_p (port, opts, 1);
+              set_port_square_brackets_p (port, opts, 1);
+              set_port_keyword_style (port, opts, KEYWORD_STYLE_HASH_PREFIX);
+              set_port_hungry_eol_escapes_p (port, opts, 1);
+            }
           else if (0 == strcmp ("fold-case", name))
             set_port_case_insensitive_p (port, opts, 1);
           else if (0 == strcmp ("no-fold-case", name))
@@ -1532,7 +1556,8 @@ scm_read_extended_symbol (scm_t_wchar chr, SCM port)
   size_t len = 0;
   SCM buf = scm_i_make_string (1024, NULL, 0);
 
-  buf = scm_i_string_start_writing (buf);
+  /* No need to scm_i_string_start_writing (), as the string isn't
+     visible to any other thread.  */
 
   while ((chr = scm_getc (port)) != EOF)
     {
@@ -1596,16 +1621,13 @@ scm_read_extended_symbol (scm_t_wchar chr, SCM port)
 	{
 	  SCM addy;
 
-	  scm_i_string_stop_writing ();
 	  addy = scm_i_make_string (1024, NULL, 0);
 	  buf = scm_string_append (scm_list_2 (buf, addy));
 	  len = 0;
-	  buf = scm_i_string_start_writing (buf);
 	}
     }
 
  done:
-  scm_i_string_stop_writing ();
   if (chr == EOF)
     scm_i_input_error ("scm_read_extended_symbol", port,
                        "end of file while reading symbol", SCM_EOL);
@@ -1625,8 +1647,8 @@ scm_read_sharp_extension (int chr, SCM port, scm_t_read_opts *opts)
   proc = scm_get_hash_procedure (chr);
   if (scm_is_true (scm_procedure_p (proc)))
     {
-      long line = SCM_LINUM (port);
-      int column = SCM_COL (port) - 2;
+      long line = scm_to_long (scm_port_line (port));
+      int column = scm_to_int (scm_port_column (port)) - 2;
       SCM got;
 
       got = scm_call_2 (proc, SCM_MAKE_CHAR (chr), port);
@@ -1681,29 +1703,10 @@ scm_read_sharp (scm_t_wchar chr, SCM port, scm_t_read_opts *opts,
     case '0': case '1': case '2': case '3': case '4':
     case '5': case '6': case '7': case '8': case '9':
     case '@':
-#if SCM_ENABLE_DEPRECATED
-      /* See below for 'i' and 'e'. */
-    case 'a':
-    case 'y':
-    case 'h':
-    case 'l':
-#endif
       return (scm_read_array (chr, port, opts, line, column));
 
     case 'i':
     case 'e':
-#if SCM_ENABLE_DEPRECATED
-      {
-	/* When next char is '(', it really is an old-style
-	   uniform array. */
-	scm_t_wchar next_c = scm_getc (port);
-	if (next_c != EOF)
-	  scm_ungetc (next_c, port);
-	if (next_c == '(')
-	  return scm_read_array (chr, port, opts, line, column);
-	/* Fall through. */
-      }
-#endif
     case 'b':
     case 'B':
     case 'o':
@@ -1797,8 +1800,8 @@ read_inner_expression (SCM port, scm_t_read_opts *opts)
                  be part of an unescaped symbol.  We might as well do
                  something useful with it, so we adopt Kawa's convention:
                  [...] => ($bracket-list$ ...) */
-              long line = SCM_LINUM (port);
-              int column = SCM_COL (port) - 1;
+              long line = scm_to_long (scm_port_line (port));
+              int column = scm_to_int (scm_port_column (port)) - 1;
               return maybe_annotate_source
                 (scm_cons (sym_bracket_list, scm_read_sexp (chr, port, opts)),
                  port, opts, line, column);
@@ -1820,8 +1823,8 @@ read_inner_expression (SCM port, scm_t_read_opts *opts)
 	  return (scm_read_quote (chr, port, opts));
 	case '#':
 	  {
-            long line  = SCM_LINUM (port);
-            int column = SCM_COL (port) - 1;
+            long line = scm_to_long (scm_port_line (port));
+            int column = scm_to_int (scm_port_column (port)) - 1;
 	    SCM result = scm_read_sharp (chr, port, opts, line, column);
 	    if (scm_is_eq (result, SCM_UNSPECIFIED))
 	      /* We read a comment or some such.  */
@@ -1885,8 +1888,8 @@ scm_read_expression (SCM port, scm_t_read_opts *opts)
           if (c == EOF)
             return SCM_EOF_VAL;
           scm_ungetc (c, port);
-          line = SCM_LINUM (port);
-          column = SCM_COL (port);
+          line = scm_to_long (scm_port_line (port));
+          column = scm_to_int (scm_port_column (port));
         }
 
       expr = read_inner_expression (port, opts);
@@ -2046,6 +2049,15 @@ scm_get_hash_procedure (int c)
     }
 }
 
+static int
+is_encoding_char (char c)
+{
+  if (c >= 'a' && c <= 'z') return 1;
+  if (c >= 'A' && c <= 'Z') return 1;
+  if (c >= '0' && c <= '9') return 1;
+  return strchr ("_-.:/,+=()", c) != NULL;
+}
+
 /* Maximum size of an encoding name.  This is a bit more than the
    longest name listed at
    <http://www.iana.org/assignments/character-sets> ("ISO-2022-JP-2", 13
@@ -2064,27 +2076,23 @@ char *
 scm_i_scan_for_encoding (SCM port)
 {
   scm_t_port *pt;
+  SCM buf;
   char header[SCM_ENCODING_SEARCH_SIZE+1];
-  size_t bytes_read, encoding_length, i;
+  size_t cur, bytes_read, encoding_length, i;
   char *encoding = NULL;
   char *pos, *encoding_start;
   int in_comment;
 
-  pt = SCM_PTAB_ENTRY (port);
-
-  if (pt->rw_active == SCM_PORT_WRITE)
-    scm_flush (port);
+  pt = SCM_PORT (port);
+  buf = pt->read_buf;
 
   if (pt->rw_random)
-    pt->rw_active = SCM_PORT_READ;
+    scm_flush (port);
 
-  if (pt->read_pos == pt->read_end)
+  if (scm_port_buffer_can_take (buf, &cur) == 0)
     {
       /* We can use the read buffer, and thus avoid a seek. */
-      if (scm_fill_input (port) == EOF)
-        return NULL;
-
-      bytes_read = pt->read_end - pt->read_pos;
+      buf = scm_fill_input (port, 0, &cur, &bytes_read);
       if (bytes_read > SCM_ENCODING_SEARCH_SIZE)
         bytes_read = SCM_ENCODING_SEARCH_SIZE;
 
@@ -2092,24 +2100,20 @@ scm_i_scan_for_encoding (SCM port)
         /* An unbuffered port -- don't scan.  */
         return NULL;
 
-      memcpy (header, pt->read_pos, bytes_read);
+      memcpy (header, scm_port_buffer_take_pointer (buf, cur), bytes_read);
       header[bytes_read] = '\0';
     }
-  else
+  else if (pt->rw_random)
     {
-      /* Try to read some bytes and then seek back.  Not all ports
-         support seeking back; and indeed some file ports (like
-         /dev/urandom) will succeed on an lseek (fd, 0, SEEK_CUR)---the
-         check performed by SCM_FPORT_FDES---but fail to seek
-         backwards.  Hence this block comes second.  We prefer to use
-         the read buffer in-place.  */
-      if (SCM_FPORTP (port) && !SCM_FDES_RANDOM_P (SCM_FPORT_FDES (port)))
-        return NULL;
-
+      /* The port is seekable.  This is OK but grubbing in the read
+         buffer is better, so this case is just a fallback.  */
       bytes_read = scm_c_read (port, header, SCM_ENCODING_SEARCH_SIZE);
       header[bytes_read] = '\0';
       scm_seek (port, scm_from_int (0), scm_from_int (SEEK_SET));
     }
+  else
+    /* No input available and not seekable; scan fails.  */
+    return NULL;
 
   /* search past "coding[:=]" */
   pos = header;
@@ -2144,8 +2148,7 @@ scm_i_scan_for_encoding (SCM port)
   i = 0;
   while (encoding_start + i - header <= SCM_ENCODING_SEARCH_SIZE
          && encoding_start + i - header < bytes_read
-	 && (isalnum ((int) encoding_start[i])
-	     || strchr ("_-.:/,+=()", encoding_start[i]) != NULL))
+	 && is_encoding_char (encoding_start[i]))
     i++;
 
   encoding_length = i;
@@ -2153,8 +2156,6 @@ scm_i_scan_for_encoding (SCM port)
     return NULL;
 
   encoding = scm_gc_strndup (encoding_start, encoding_length, "encoding");
-  for (i = 0; i < encoding_length; i++)
-    encoding[i] = toupper ((int) encoding[i]);
 
   /* push backwards to make sure we were in a comment */
   in_comment = 0;
@@ -2210,7 +2211,7 @@ SCM_DEFINE (scm_file_encoding, "file-encoding", 1, 0, 0,
     return SCM_BOOL_F;
   else
     {
-      s_enc = scm_from_locale_string (enc);
+      s_enc = scm_string_upcase (scm_from_locale_string (enc));
       return s_enc;
     }
 
@@ -2264,6 +2265,7 @@ set_port_read_option (SCM port, int option, int new_value)
   unsigned int read_options;
 
   new_value &= READ_OPTION_MASK;
+
   scm_read_options = scm_i_port_property (port, sym_port_read_options);
   if (scm_is_unsigned_integer (scm_read_options, 0, READ_OPTIONS_MAX_VALUE))
     read_options = scm_to_uint (scm_read_options);
@@ -2302,6 +2304,30 @@ set_port_curly_infix_p (SCM port, scm_t_read_opts *opts, int value)
   set_port_read_option (port, READ_OPTION_CURLY_INFIX_P, value);
 }
 
+/* Set OPTS and PORT's r6rs_hex_escapes_p option according to VALUE. */
+static void
+set_port_r6rs_hex_escapes_p (SCM port, scm_t_read_opts *opts, int value)
+{
+  value = !!value;
+  opts->r6rs_escapes_p = value;
+  set_port_read_option (port, READ_OPTION_R6RS_ESCAPES_P, value);
+}
+
+static void
+set_port_hungry_eol_escapes_p (SCM port, scm_t_read_opts *opts, int value)
+{
+  value = !!value;
+  opts->hungry_eol_escapes_p = value;
+  set_port_read_option (port, READ_OPTION_HUNGRY_EOL_ESCAPES_P, value);
+}
+
+static void
+set_port_keyword_style (SCM port, scm_t_read_opts *opts, enum t_keyword_style value)
+{
+  opts->keyword_style = value;
+  set_port_read_option (port, READ_OPTION_KEYWORD_STYLE, value);
+}
+
 /* Initialize OPTS based on PORT's read options and the global read
    options. */
 static void
