@@ -1,4 +1,5 @@
-/* Copyright (C) 1995,1996,1998,2000,2001, 2004, 2006, 2008, 2009, 2010, 2011, 2012 Free Software Foundation, Inc.
+/* Copyright (C) 1995, 1996, 1998, 2000, 2001, 2004, 2006,
+ *   2008-2016 Free Software Foundation, Inc.
  * 
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -29,13 +30,15 @@
 #include <uninorm.h>
 #include <unistr.h>
 #include <uniconv.h>
+#include <c-strcase.h>
 
 #include "striconveh.h"
 
 #include "libguile/_scm.h"
 #include "libguile/chars.h"
-#include "libguile/root.h"
 #include "libguile/strings.h"
+#include "libguile/ports.h"
+#include "libguile/ports-internal.h"
 #include "libguile/error.h"
 #include "libguile/generalized-vectors.h"
 #include "libguile/deprecation.h"
@@ -47,41 +50,42 @@
 /* {Strings}
  */
 
+SCM_SYMBOL (sym_UTF_8, "UTF-8");
+SCM_SYMBOL (sym_ISO_8859_1, "ISO-8859-1");
+SCM_SYMBOL (sym_error, "error");
 
-/* Stringbufs 
- *
- * XXX - keeping an accurate refcount during GC seems to be quite
- * tricky, so we just keep score of whether a stringbuf might be
- * shared, not whether it definitely is.  
- *
- * The scheme I (mvo) tried to keep an accurate reference count would
- * recount all strings that point to a stringbuf during the mark-phase
- * of the GC.  This was done since one cannot access the stringbuf of
- * a string when that string is freed (in order to decrease the
- * reference count).  The memory of the stringbuf might have been
- * reused already for something completely different.
- *
- * This recounted worked for a small number of threads beating on
- * cow-strings, but it failed randomly with more than 10 threads, say.
- * I couldn't figure out what went wrong, so I used the conservative
- * approach implemented below.
- *
- * There are 2 storage strategies for stringbufs: 8-bit and wide.  8-bit
- * strings are ISO-8859-1-encoded strings; wide strings are 32-bit (UCS-4)
- * strings.
- */
+/* A stringbuf is a linear buffer of characters.  Every string has a
+   stringbuf.  Strings may reference just a slice of a stringbuf; that's
+   often the case for strings made by the "substring" function.
+
+   Stringbufs may hold either 8-bit characters or 32-bit characters.  In
+   either case the characters are Unicode codepoints.  "Narrow"
+   stringbufs thus have the ISO-8859-1 (Latin-1) encoding, and "wide"
+   stringbufs have the UTF-32 (UCS-4) encoding.
+
+   By default, stringbufs are immutable.  This enables an O(1)
+   "substring" operation with no synchronization.  A string-set! will
+   first ensure that the string's stringbuf is mutable, copying the
+   stringbuf if necessary.  This is therefore a copy-on-write
+   representation.  However, taking a substring of a mutable stringbuf
+   is an O(n) operation as it has to create a new immutable stringbuf.
+   There are also mutation-sharing substrings as well.  */
 
 /* The size in words of the stringbuf header (type tag + size).  */
 #define STRINGBUF_HEADER_SIZE   2U
 
 #define STRINGBUF_HEADER_BYTES  (STRINGBUF_HEADER_SIZE * sizeof (SCM))
 
-#define STRINGBUF_F_SHARED      SCM_I_STRINGBUF_F_SHARED
 #define STRINGBUF_F_WIDE        SCM_I_STRINGBUF_F_WIDE
+#define STRINGBUF_F_MUTABLE     SCM_I_STRINGBUF_F_MUTABLE
 
 #define STRINGBUF_TAG           scm_tc7_stringbuf
-#define STRINGBUF_SHARED(buf)   (SCM_CELL_WORD_0(buf) & STRINGBUF_F_SHARED)
 #define STRINGBUF_WIDE(buf)     (SCM_CELL_WORD_0(buf) & STRINGBUF_F_WIDE)
+#define STRINGBUF_MUTABLE(buf)  (SCM_CELL_WORD_0(buf) & STRINGBUF_F_MUTABLE)
+
+
+#define STRINGBUF_SET_MUTABLE(buf) \
+  SCM_SET_CELL_WORD_0 (buf, SCM_CELL_WORD_0 (buf) | STRINGBUF_F_MUTABLE)
 
 #define STRINGBUF_CONTENTS(buf) ((void *)				\
                                  SCM_CELL_OBJECT_LOC (buf,		\
@@ -91,16 +95,6 @@
 
 #define STRINGBUF_LENGTH(buf)   (SCM_CELL_WORD_1 (buf))
 
-#define SET_STRINGBUF_SHARED(buf)					\
-  do									\
-    {									\
-      /* Don't modify BUF if it's already marked as shared since it might be \
-	 a read-only, statically allocated stringbuf.  */		\
-      if (SCM_LIKELY (!STRINGBUF_SHARED (buf)))				\
-	SCM_SET_CELL_WORD_0 ((buf), SCM_CELL_WORD_0 (buf) | STRINGBUF_F_SHARED); \
-    }									\
-  while (0)
-
 #ifdef SCM_STRING_LENGTH_HISTOGRAM
 static size_t lenhist[1001];
 #endif
@@ -126,7 +120,7 @@ make_stringbuf (size_t len)
     lenhist[1000]++;
 #endif
 
-  buf = PTR2SCM (scm_gc_malloc_pointerless (STRINGBUF_HEADER_BYTES + len + 1,
+  buf = SCM_PACK_POINTER (scm_gc_malloc_pointerless (STRINGBUF_HEADER_BYTES + len + 1,
 					    "string"));
 
   SCM_SET_CELL_TYPE (buf, STRINGBUF_TAG);
@@ -153,7 +147,7 @@ make_wide_stringbuf (size_t len)
 #endif
 
   raw_len = (len + 1) * sizeof (scm_t_wchar);
-  buf = PTR2SCM (scm_gc_malloc_pointerless (STRINGBUF_HEADER_BYTES + raw_len,
+  buf = SCM_PACK_POINTER (scm_gc_malloc_pointerless (STRINGBUF_HEADER_BYTES + raw_len,
 					    "string"));
 
   SCM_SET_CELL_TYPE (buf, STRINGBUF_TAG | STRINGBUF_F_WIDE);
@@ -225,8 +219,6 @@ narrow_stringbuf (SCM buf)
   return new_buf;
 }
 
-scm_i_pthread_mutex_t stringbuf_write_mutex = SCM_I_PTHREAD_MUTEX_INITIALIZER;
-
 
 /* Copy-on-write strings.
  */
@@ -240,7 +232,7 @@ scm_i_pthread_mutex_t stringbuf_write_mutex = SCM_I_PTHREAD_MUTEX_INITIALIZER;
 #define SET_STRING_STRINGBUF(str,buf) (SCM_SET_CELL_OBJECT_1(str,buf))
 #define SET_STRING_START(str,start) (SCM_SET_CELL_WORD_2(str,start))
 
-#define IS_STRING(str)        (SCM_NIMP(str) && SCM_TYP7(str) == STRING_TAG)
+#define IS_STRING(str)        (SCM_HAS_TYP7 (str, STRING_TAG))
 
 /* Read-only strings.
  */
@@ -258,8 +250,26 @@ scm_i_pthread_mutex_t stringbuf_write_mutex = SCM_I_PTHREAD_MUTEX_INITIALIZER;
 
 #define IS_SH_STRING(str)   (SCM_CELL_TYPE(str)==SH_STRING_TAG)
 
+void
+scm_i_print_stringbuf (SCM exp, SCM port, scm_print_state *pstate) 
+{
+  SCM str = scm_double_cell (STRING_TAG, SCM_UNPACK(exp), 0,
+                             STRINGBUF_LENGTH (exp));
+  scm_puts ("#<stringbuf ", port);
+  scm_iprin1 (str, port, pstate);
+  scm_puts (">", port);
+}
+
 SCM scm_nullstr;
 
+static SCM null_stringbuf;
+
+static void
+init_null_stringbuf (void)
+{
+  null_stringbuf = make_stringbuf (0);
+}
+
 /* Create a scheme string with space for LEN 8-bit Latin-1-encoded
    characters.  CHARSP, if not NULL, will be set to location of the
    char array.  If READ_ONLY_P, the returned string is read-only;
@@ -267,17 +277,13 @@ SCM scm_nullstr;
 SCM
 scm_i_make_string (size_t len, char **charsp, int read_only_p)
 {
-  static SCM null_stringbuf = SCM_BOOL_F;
   SCM buf;
   SCM res;
 
   if (len == 0)
     {
-      if (SCM_UNLIKELY (scm_is_false (null_stringbuf)))
-        {
-          null_stringbuf = make_stringbuf (0);
-          SET_STRINGBUF_SHARED (null_stringbuf);
-        }
+      static scm_i_pthread_once_t once = SCM_I_PTHREAD_ONCE_INIT;
+      scm_i_pthread_once (&once, init_null_stringbuf);
       buf = null_stringbuf;
     }
   else
@@ -331,77 +337,109 @@ get_str_buf_start (SCM *str, SCM *buf, size_t *start)
   *buf = STRING_STRINGBUF (*str);
 }
 
+static SCM
+substring_with_immutable_stringbuf (SCM str, size_t start, size_t end,
+                                    int force_copy_p, int read_only_p)
+{
+  SCM buf;
+  size_t str_start, len;
+  scm_t_bits tag = read_only_p ? RO_STRING_TAG : STRING_TAG;
+
+  get_str_buf_start (&str, &buf, &str_start);
+  len = end - start;
+  start += str_start;
+
+  if (len == 0)
+    return scm_i_make_string (0, NULL, read_only_p);
+  else if (!force_copy_p && SCM_LIKELY (!STRINGBUF_MUTABLE (buf)))
+    return scm_double_cell (tag, SCM_UNPACK (buf), start, len);
+  else
+    {
+      SCM new_buf, new_str;
+
+      if (STRINGBUF_WIDE (buf))
+        {
+          new_buf = make_wide_stringbuf (len);
+          u32_cpy ((scm_t_uint32 *) STRINGBUF_WIDE_CHARS (new_buf),
+                   (scm_t_uint32 *) (STRINGBUF_WIDE_CHARS (buf) + start), len);
+          new_str = scm_double_cell (tag, SCM_UNPACK (new_buf), 0, len);
+          scm_i_try_narrow_string (new_str);
+        }
+      else
+        {
+          new_buf = make_stringbuf (len);
+          memcpy (STRINGBUF_CHARS (new_buf),
+                  STRINGBUF_CHARS (buf) + start, len);
+          new_str = scm_double_cell (tag, SCM_UNPACK (new_buf), 0, len);
+        }
+
+      return new_str;
+    }
+}
+
 SCM
 scm_i_substring (SCM str, size_t start, size_t end)
 {
-  if (start == end)
-    return scm_i_make_string (0, NULL, 0);
-  else
-    {
-      SCM buf;
-      size_t str_start;
-      get_str_buf_start (&str, &buf, &str_start);
-      scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-      SET_STRINGBUF_SHARED (buf);
-      scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
-      return scm_double_cell (STRING_TAG, SCM_UNPACK(buf),
-                              (scm_t_bits)str_start + start,
-                              (scm_t_bits) end - start);
-    }
+  return substring_with_immutable_stringbuf (str, start, end, 0, 0);
 }
 
 SCM
 scm_i_substring_read_only (SCM str, size_t start, size_t end)
 {
-  if (start == end)
-    return scm_i_make_string (0, NULL, 1);
-  else
-    {
-      SCM buf;
-      size_t str_start;
-      get_str_buf_start (&str, &buf, &str_start);
-      scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-      SET_STRINGBUF_SHARED (buf);
-      scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
-      return scm_double_cell (RO_STRING_TAG, SCM_UNPACK(buf),
-                              (scm_t_bits)str_start + start,
-                              (scm_t_bits) end - start);
-    }
+  return substring_with_immutable_stringbuf (str, start, end, 0, 1);
 }
 
 SCM
 scm_i_substring_copy (SCM str, size_t start, size_t end)
 {
-  if (start == end)
-    return scm_i_make_string (0, NULL, 0);
-  else
+  return substring_with_immutable_stringbuf (str, start, end, 1, 0);
+}
+
+static void
+scm_i_string_ensure_mutable_x (SCM str)
+{
+  SCM buf;
+
+  if (IS_SH_STRING (str))
     {
-      size_t len = end - start;
-      SCM buf, my_buf, substr;
-      size_t str_start;
-      int wide = 0;
-      get_str_buf_start (&str, &buf, &str_start);
-      if (scm_i_is_narrow_string (str))
-        {
-          my_buf = make_stringbuf (len);
-          memcpy (STRINGBUF_CHARS (my_buf),
-                  STRINGBUF_CHARS (buf) + str_start + start, len);
-        }
-      else
-        {
-          my_buf = make_wide_stringbuf (len);
-          u32_cpy ((scm_t_uint32 *) STRINGBUF_WIDE_CHARS (my_buf),
-                   (scm_t_uint32 *) (STRINGBUF_WIDE_CHARS (buf) + str_start 
-                                     + start), len);
-          wide = 1;
-        }
-      scm_remember_upto_here_1 (buf);
-      substr = scm_double_cell (STRING_TAG, SCM_UNPACK (my_buf),
-                                (scm_t_bits) 0, (scm_t_bits) len);
-      if (wide)
-        scm_i_try_narrow_string (substr);
-      return substr;
+      /* Shared-mutation strings always have mutable stringbufs.  */
+      buf = STRING_STRINGBUF (SH_STRING_STRING (str));
+      if (!STRINGBUF_MUTABLE (buf))
+        abort ();
+      return;
     }
+
+  if (IS_RO_STRING (str))
+    scm_misc_error (NULL, "string is read-only: ~s", scm_list_1 (str));
+
+  buf = STRING_STRINGBUF (str);
+
+  if (STRINGBUF_MUTABLE (buf))
+    return;
+
+  /* Otherwise copy and mark the fresh stringbuf as mutable.  Note that
+     we copy the whole stringbuf so that the start/len offsets from the
+     original string keep working, so that concurrent accessors on this
+     string don't see things in an inconsistent state.  */
+  {
+    SCM new_buf;
+    size_t len = STRINGBUF_LENGTH (buf);
+
+    if (STRINGBUF_WIDE (buf))
+      {
+        new_buf = make_wide_stringbuf (len);
+        u32_cpy ((scm_t_uint32 *) STRINGBUF_WIDE_CHARS (new_buf),
+                 (scm_t_uint32 *) STRINGBUF_WIDE_CHARS (buf), len);
+      }
+    else
+      {
+        new_buf = make_stringbuf (len);
+        memcpy (STRINGBUF_CHARS (new_buf), STRINGBUF_CHARS (buf), len);
+      }
+
+    STRINGBUF_SET_MUTABLE (new_buf);
+    SET_STRING_STRINGBUF (str, new_buf);
+  }
 }
 
 SCM
@@ -411,6 +449,8 @@ scm_i_substring_shared (SCM str, size_t start, size_t end)
     return str;
   else if (start == end)
     return scm_i_make_string (0, NULL, 0);
+  else if (IS_RO_STRING (str))
+    return scm_i_substring_read_only (str, start, end);
   else
     {
       size_t len = end - start;
@@ -419,6 +459,9 @@ scm_i_substring_shared (SCM str, size_t start, size_t end)
 	  start += STRING_START (str);
 	  str = SH_STRING_STRING (str);
 	}
+
+      scm_i_string_ensure_mutable_x (str);
+
       return scm_double_cell (SH_STRING_TAG, SCM_UNPACK(str),
 			      (scm_t_bits)start, (scm_t_bits) len);
     }
@@ -464,6 +507,12 @@ scm_i_string_length (SCM str)
   return STRING_LENGTH (str);
 }
 
+int
+scm_i_string_is_mutable (SCM str)
+{
+  return !IS_RO_STRING (str);
+}
+
 /* True if the string is 'narrow', meaning it has a 8-bit Latin-1
    encoding.  False if it is 'wide', having a 32-bit UCS-4
    encoding.  */
@@ -540,60 +589,13 @@ scm_i_string_wide_chars (SCM str)
                     scm_list_1 (str));
 }
 
-/* If the buffer in ORIG_STR is shared, copy ORIG_STR's characters to
+/* If the buffer in ORIG_STR is immutable, copy ORIG_STR's characters to
    a new string buffer, so that it can be modified without modifying
-   other strings.  Also, lock the string mutex.  Later, one must call
-   scm_i_string_stop_writing to unlock the mutex.  */
+   other strings.  */
 SCM
 scm_i_string_start_writing (SCM orig_str)
 {
-  SCM buf, str = orig_str;
-  size_t start;
-
-  get_str_buf_start (&str, &buf, &start);
-  if (IS_RO_STRING (str))
-    scm_misc_error (NULL, "string is read-only: ~s", scm_list_1 (orig_str));
-
-  scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-  if (STRINGBUF_SHARED (buf))
-    {
-      /* Clone the stringbuf.  */
-      size_t len = STRING_LENGTH (str);
-      SCM new_buf;
-
-      scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
-
-      if (scm_i_is_narrow_string (str))
-        {
-          new_buf = make_stringbuf (len);
-          memcpy (STRINGBUF_CHARS (new_buf),
-                  STRINGBUF_CHARS (buf) + STRING_START (str), len);
-
-        }
-      else
-        {
-          new_buf = make_wide_stringbuf (len);
-          u32_cpy ((scm_t_uint32 *) STRINGBUF_WIDE_CHARS (new_buf),
-                   (scm_t_uint32 *) (STRINGBUF_WIDE_CHARS (buf) 
-                                     + STRING_START (str)), len);
-        }
-
-      SET_STRING_STRINGBUF (str, new_buf);
-      start -= STRING_START (str);
-
-      /* FIXME: The following operations are not atomic, so other threads
-	 looking at STR may see an inconsistent state.  Nevertheless it can't
-	 hurt much since (i) accessing STR while it is being mutated can't
-	 yield a crash, and (ii) concurrent accesses to STR should be
-	 protected by a mutex at the application level.  The latter may not
-	 apply when STR != ORIG_STR, though.  */
-      SET_STRING_START (str, 0);
-      SET_STRING_STRINGBUF (str, new_buf);
-
-      buf = new_buf;
-
-      scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-    }
+  scm_i_string_ensure_mutable_x (orig_str);
   return orig_str;
 }
 
@@ -633,7 +635,6 @@ scm_i_string_writable_wide_chars (SCM str)
 void
 scm_i_string_stop_writing (void)
 {
-  scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
 }
 
 /* Return the Xth character of STR as a UCS-4 codepoint.  */
@@ -740,42 +741,10 @@ scm_i_make_symbol (SCM name, scm_t_bits flags,
 		   unsigned long hash, SCM props)
 {
   SCM buf;
-  size_t start = STRING_START (name);
   size_t length = STRING_LENGTH (name);
 
-  if (IS_SH_STRING (name))
-    {
-      name = SH_STRING_STRING (name);
-      start += STRING_START (name);
-    }
+  name = scm_i_substring_copy (name, 0, length);
   buf = STRING_STRINGBUF (name);
-
-  if (start == 0 && length == STRINGBUF_LENGTH (buf))
-    {
-      /* reuse buf. */
-      scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-      SET_STRINGBUF_SHARED (buf);
-      scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
-    }
-  else
-    {
-      /* make new buf. */
-      if (scm_i_is_narrow_string (name))
-        {
-          SCM new_buf = make_stringbuf (length);
-          memcpy (STRINGBUF_CHARS (new_buf),
-                  STRINGBUF_CHARS (buf) + start, length);
-          buf = new_buf;
-        }
-      else
-        {
-          SCM new_buf = make_wide_stringbuf (length);
-          u32_cpy ((scm_t_uint32 *) STRINGBUF_WIDE_CHARS (new_buf),
-                   (scm_t_uint32 *) STRINGBUF_WIDE_CHARS (buf) + start,
-                   length);
-          buf = new_buf;
-        }
-    }
   return scm_double_cell (scm_tc7_symbol | flags, SCM_UNPACK (buf),
 			  (scm_t_bits) hash, SCM_UNPACK (props));
 }
@@ -854,9 +823,6 @@ SCM
 scm_i_symbol_substring (SCM sym, size_t start, size_t end)
 {
   SCM buf = SYMBOL_STRINGBUF (sym);
-  scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-  SET_STRINGBUF_SHARED (buf);
-  scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
   return scm_double_cell (RO_STRING_TAG, SCM_UNPACK (buf),
 			  (scm_t_bits)start, (scm_t_bits) end - start);
 }
@@ -893,8 +859,8 @@ SCM_DEFINE (scm_sys_string_dump, "%string-dump", 1, 0, 0, (SCM str),
             "A new string containing this string's stringbuf's characters\n"
             "@item stringbuf-length\n"
             "The number of characters in this stringbuf\n"
-            "@item stringbuf-shared\n"
-            "@code{#t} if this stringbuf is shared\n"
+            "@item stringbuf-mutable\n"
+            "@code{#t} if this stringbuf is mutable\n"
             "@item stringbuf-wide\n"
             "@code{#t} if this stringbuf's characters are stored in a\n"
             "32-bit buffer, or @code{#f} if they are stored in an 8-bit\n"
@@ -956,11 +922,11 @@ SCM_DEFINE (scm_sys_string_dump, "%string-dump", 1, 0, 0, (SCM str),
     }
   e7 = scm_cons (scm_from_latin1_symbol ("stringbuf-length"), 
                  scm_from_size_t (STRINGBUF_LENGTH (buf)));
-  if (STRINGBUF_SHARED (buf))
-    e8 = scm_cons (scm_from_latin1_symbol ("stringbuf-shared"), 
+  if (STRINGBUF_MUTABLE (buf))
+    e8 = scm_cons (scm_from_latin1_symbol ("stringbuf-mutable"),
                    SCM_BOOL_T);
   else
-    e8 = scm_cons (scm_from_latin1_symbol ("stringbuf-shared"), 
+    e8 = scm_cons (scm_from_latin1_symbol ("stringbuf-mutable"),
                    SCM_BOOL_F);
   if (STRINGBUF_WIDE (buf))
     e9 = scm_cons (scm_from_latin1_symbol ("stringbuf-wide"),
@@ -987,8 +953,8 @@ SCM_DEFINE (scm_sys_symbol_dump, "%symbol-dump", 1, 0, 0, (SCM sym),
             "A new string containing this symbols's stringbuf's characters\n"
             "@item stringbuf-length\n"
             "The number of characters in this stringbuf\n"
-            "@item stringbuf-shared\n"
-            "@code{#t} if this stringbuf is shared\n"
+            "@item stringbuf-mutable\n"
+            "@code{#t} if this stringbuf is mutable\n"
             "@item stringbuf-wide\n"
             "@code{#t} if this stringbuf's characters are stored in a\n"
             "32-bit buffer, or @code{#f} if they are stored in an 8-bit\n"
@@ -1029,11 +995,11 @@ SCM_DEFINE (scm_sys_symbol_dump, "%symbol-dump", 1, 0, 0, (SCM sym),
     }
   e5 = scm_cons (scm_from_latin1_symbol ("stringbuf-length"), 
                  scm_from_size_t (STRINGBUF_LENGTH (buf)));
-  if (STRINGBUF_SHARED (buf))
-    e6 = scm_cons (scm_from_latin1_symbol ("stringbuf-shared"), 
+  if (STRINGBUF_MUTABLE (buf))
+    e6 = scm_cons (scm_from_latin1_symbol ("stringbuf-mutable"),
                    SCM_BOOL_T);
   else
-    e6 = scm_cons (scm_from_latin1_symbol ("stringbuf-shared"), 
+    e6 = scm_cons (scm_from_latin1_symbol ("stringbuf-mutable"),
                    SCM_BOOL_F);
   if (STRINGBUF_WIDE (buf))
     e7 = scm_cons (scm_from_latin1_symbol ("stringbuf-wide"),
@@ -1117,7 +1083,6 @@ SCM_DEFINE (scm_string, "string", 0, 0, 1,
       char *buf;
 
       result = scm_i_make_string (len, NULL, 0);
-      result = scm_i_string_start_writing (result);
       buf = scm_i_string_writable_chars (result);
       while (len > 0 && scm_is_pair (rest))
         {
@@ -1134,7 +1099,6 @@ SCM_DEFINE (scm_string, "string", 0, 0, 1,
       scm_t_wchar *buf;
 
       result = scm_i_make_wide_string (len, NULL, 0);
-      result = scm_i_string_start_writing (result);
       buf = scm_i_string_writable_wide_chars (result);
       while (len > 0 && scm_is_pair (rest))
         {
@@ -1146,7 +1110,6 @@ SCM_DEFINE (scm_string, "string", 0, 0, 1,
           scm_remember_upto_here_1 (elt);
         }
     }
-  scm_i_string_stop_writing ();
 
   if (len > 0)
     scm_misc_error (NULL, "list changed while constructing string", SCM_EOL);
@@ -1165,7 +1128,12 @@ SCM_DEFINE (scm_make_string, "make-string", 1, 1, 0,
 	    "of the string are all set to @code{#\nul}.")
 #define FUNC_NAME s_scm_make_string
 {
-  return scm_c_make_string (scm_to_size_t (k), chr);
+  SCM ret = scm_c_make_string (scm_to_size_t (k), chr);
+  /* Given that make-string is mostly used by Scheme to prepare a
+     mutable string buffer, let's go ahead and mark this as mutable to
+     avoid a copy when this buffer is next written to.  */
+  STRINGBUF_SET_MUTABLE (STRING_STRINGBUF (ret));
+  return ret;
 }
 #undef FUNC_NAME
 
@@ -1183,10 +1151,8 @@ scm_c_make_string (size_t len, SCM chr)
   else
     {
       SCM_VALIDATE_CHAR (0, chr);
-      res = scm_i_string_start_writing (res);
       for (p = 0; p < len; p++)
         scm_i_string_set_x (res, p, SCM_CHAR (chr));
-      scm_i_string_stop_writing ();
     }
 
   return res;
@@ -1502,6 +1468,23 @@ scm_decoding_error (const char *subr, int err, const char *message, SCM port)
 
 /* String conversion to/from C.  */
 
+static void
+decoding_error (const char *func_name, int errno_save,
+                const char *str, size_t len)
+{
+  /* Raise an error and pass the raw C string as a bytevector to the `throw'
+     handler.  */
+  SCM bv;
+  signed char *buf;
+
+  buf = scm_gc_malloc_pointerless (len, "bytevector");
+  memcpy (buf, str, len);
+  bv = scm_c_take_gc_bytevector (buf, len, SCM_BOOL_F);
+
+  scm_decoding_error (func_name, errno_save,
+                      "input locale conversion error", bv);
+}
+
 SCM
 scm_from_stringn (const char *str, size_t len, const char *encoding,
                   scm_t_string_failed_conversion_handler handler)
@@ -1517,14 +1500,11 @@ scm_from_stringn (const char *str, size_t len, const char *encoding,
   if (len == (size_t) -1)
     len = strlen (str);
 
-  if (encoding == NULL || len == 0)
-    {
-      /* If encoding is null (or the string is empty), use Latin-1.  */
-      char *buf;
-      res = scm_i_make_string (len, &buf, 0);
-      memcpy (buf, str, len);
-      return res;
-    }
+  if (c_strcasecmp (encoding, "ISO-8859-1") == 0 || len == 0)
+    return scm_from_latin1_stringn (str, len);
+  else if (c_strcasecmp (encoding, "UTF-8") == 0
+           && handler == SCM_FAILED_CONVERSION_ERROR)
+    return scm_from_utf8_stringn (str, len);
 
   u32len = 0;
   u32 = (scm_t_wchar *) u32_conv_from_encoding (encoding,
@@ -1535,19 +1515,7 @@ scm_from_stringn (const char *str, size_t len, const char *encoding,
                                                 NULL, &u32len);
 
   if (SCM_UNLIKELY (u32 == NULL))
-    {
-      /* Raise an error and pass the raw C string as a bytevector to the `throw'
-	 handler.  */
-      SCM bv;
-      signed char *buf;
-
-      buf = scm_gc_malloc_pointerless (len, "bytevector");
-      memcpy (buf, str, len);
-      bv = scm_c_take_gc_bytevector (buf, len);
-
-      scm_decoding_error (__func__, errno,
-			  "input locale conversion error", bv);
-    }
+    decoding_error (__func__, errno, str, len);
 
   i = 0;
   while (i < u32len)
@@ -1583,11 +1551,18 @@ scm_from_locale_string (const char *str)
   return scm_from_locale_stringn (str, -1);
 }
 
+scm_t_string_failed_conversion_handler
+scm_i_default_string_failed_conversion_handler (void)
+{
+  return scm_i_string_failed_conversion_handler
+    (scm_i_default_port_conversion_strategy ());
+}
+
 SCM
 scm_from_locale_stringn (const char *str, size_t len)
 {
   return scm_from_stringn (str, len, locale_charset (),
-                           scm_i_default_port_conversion_handler ());
+                           scm_i_default_string_failed_conversion_handler ());
 }
 
 SCM
@@ -1621,7 +1596,81 @@ scm_from_utf8_string (const char *str)
 SCM
 scm_from_utf8_stringn (const char *str, size_t len)
 {
-  return scm_from_stringn (str, len, "UTF-8", SCM_FAILED_CONVERSION_ERROR);
+  size_t i, char_len;
+  const scm_t_uint8 *ustr = (const scm_t_uint8 *) str;
+  int ascii = 1, narrow = 1;
+  SCM res;
+
+  if (len == (size_t) -1)
+    len = strlen (str);
+
+  i = 0;
+  char_len = 0;
+
+  while (i < len)
+    {
+      if (ustr[i] <= 127)
+        {
+          char_len++;
+          i++;
+        }
+      else
+        {
+          ucs4_t c;
+          int nbytes;
+
+          ascii = 0;
+
+          nbytes = u8_mbtoucr (&c, ustr + i, len - i);
+
+          if (nbytes < 0)
+            /* Bad UTF-8.  */
+            decoding_error (__func__, errno, str, len);
+
+          if (c > 255)
+            narrow = 0;
+          
+          char_len++;
+          i += nbytes;
+        }
+    }
+  
+  if (ascii)
+    {
+      char *dst;
+      res = scm_i_make_string (char_len, &dst, 0);
+      memcpy (dst, str, len);
+    }
+  else if (narrow)
+    {
+      char *dst;
+      size_t j;
+      ucs4_t c;
+
+      res = scm_i_make_string (char_len, &dst, 0);
+
+      for (i = 0, j = 0; i < len; j++)
+        {
+          i += u8_mbtouc_unsafe (&c, ustr + i, len - i);
+          dst[j] = (signed char) c;
+        }
+    }
+  else
+    {
+      scm_t_wchar *dst;
+      size_t j;
+      ucs4_t c;
+
+      res = scm_i_make_wide_string (char_len, &dst, 0);
+
+      for (i = 0, j = 0; i < len; j++)
+        {
+          i += u8_mbtouc_unsafe (&c, ustr + i, len - i);
+          dst[j] = c;
+        }
+    }
+
+  return res;
 }
 
 SCM
@@ -1646,6 +1695,29 @@ scm_from_utf32_stringn (const scm_t_wchar *str, size_t len)
   return result;
 }
 
+SCM
+scm_from_port_string (const char *str, SCM port)
+{
+  return scm_from_port_stringn (str, -1, port);
+}
+
+SCM
+scm_from_port_stringn (const char *str, size_t len, SCM port)
+{
+  scm_t_port *pt = SCM_PORT (port);
+
+  if (scm_is_eq (pt->encoding, sym_ISO_8859_1))
+    return scm_from_latin1_stringn (str, len);
+  else if (scm_is_eq (pt->encoding, sym_UTF_8)
+           && (scm_is_eq (pt->conversion_strategy, sym_error)
+               || (u8_check ((uint8_t *) str, len) == NULL)))
+    return scm_from_utf8_stringn (str, len);
+  else
+    return scm_from_stringn (str, len, scm_i_symbol_chars (pt->encoding),
+                             scm_i_string_failed_conversion_handler
+                             (SCM_PORT (port)->conversion_strategy));
+}
+
 /* Create a new scheme string from the C string STR.  The memory of
    STR may be used directly as storage for the new string.  */
 /* FIXME: GC-wise, the only way to use the memory area pointed to by STR
@@ -1814,7 +1886,7 @@ scm_to_locale_stringn (SCM str, size_t *lenp)
 {
   return scm_to_stringn (str, lenp,
                          locale_charset (),
-                         scm_i_default_port_conversion_handler ());
+                         scm_i_default_string_failed_conversion_handler ());
 }
 
 char *
@@ -1930,6 +2002,38 @@ u32_u8_length_in_bytes (const scm_t_uint32 *str, size_t len)
   return ret;
 }
 
+static size_t
+utf8_length (SCM str)
+{
+  if (scm_i_is_narrow_string (str))
+    return latin1_u8_strlen ((scm_t_uint8 *) scm_i_string_chars (str),
+                             scm_i_string_length (str));
+  else
+    return u32_u8_length_in_bytes
+      ((scm_t_uint32 *) scm_i_string_wide_chars (str),
+       scm_i_string_length (str));
+}
+
+size_t
+scm_c_string_utf8_length (SCM string)
+#define FUNC_NAME "scm_c_string_utf8_length"
+{
+  SCM_VALIDATE_STRING (1, string);
+  return utf8_length (string);
+}
+#undef FUNC_NAME
+
+SCM_DEFINE (scm_string_utf8_length, "string-utf8-length", 1, 0, 0,
+	    (SCM string),
+	    "Returns the number of bytes in the UTF-8 representation of "
+            "@var{string}.")
+#define FUNC_NAME s_scm_string_utf8_length
+{
+  SCM_VALIDATE_STRING (1, string);
+  return scm_from_size_t (utf8_length (string));
+}
+#undef FUNC_NAME
+
 char *
 scm_to_utf8_stringn (SCM str, size_t *lenp)
 #define FUNC_NAME "scm_to_utf8_stringn"
@@ -2031,6 +2135,28 @@ scm_to_utf32_stringn (SCM str, size_t *lenp)
 }
 #undef FUNC_NAME
 
+char *
+scm_to_port_string (SCM str, SCM port)
+{
+  return scm_to_port_stringn (str, NULL, port);
+}
+
+char *
+scm_to_port_stringn (SCM str, size_t *lenp, SCM port)
+{
+  scm_t_port *pt = SCM_PORT (port);
+
+  if (scm_is_eq (pt->encoding, sym_ISO_8859_1)
+      && scm_is_eq (pt->conversion_strategy, sym_error))
+    return scm_to_latin1_stringn (str, lenp);
+  else if (scm_is_eq (pt->encoding, sym_UTF_8))
+    return scm_to_utf8_stringn (str, lenp);
+  else
+    return scm_to_stringn (str, lenp, scm_i_symbol_chars (pt->encoding),
+                           scm_i_string_failed_conversion_handler
+                           (SCM_PORT (port)->conversion_strategy));
+}
+
 /* Return a malloc(3)-allocated buffer containing the contents of STR encoded
    according to ENCODING.  If LENP is non-NULL, set it to the size in bytes of
    the returned buffer.  If the conversion to ENCODING fails, apply the strategy
@@ -2064,7 +2190,8 @@ scm_to_stringn (SCM str, size_t *lenp, const char *encoding,
                         "string contains #\\nul character: ~S",
                         scm_list_1 (str));
 
-  if (scm_i_is_narrow_string (str) && (encoding == NULL))
+  if (scm_i_is_narrow_string (str)
+      && c_strcasecmp (encoding, "ISO-8859-1") == 0)
     {
       /* If using native Latin-1 encoding, just copy the string
          contents.  */
@@ -2320,94 +2447,6 @@ scm_i_get_substring_spec (size_t len,
     *cend = scm_to_unsigned_integer (end, *cstart, len);
 }
 		  
-#if SCM_ENABLE_DEPRECATED
-
-/* When these definitions are removed, it becomes reasonable to use
-   read-only strings for string literals.  For that, change the reader
-   to create string literals with scm_c_substring_read_only instead of
-   with scm_c_substring_copy.
-*/
-
-int
-scm_i_deprecated_stringp (SCM str)
-{
-  scm_c_issue_deprecation_warning
-    ("SCM_STRINGP is deprecated.  Use scm_is_string instead.");
-  
-  return scm_is_string (str);
-}
-
-char *
-scm_i_deprecated_string_chars (SCM str)
-{
-  char *chars;
-
-  scm_c_issue_deprecation_warning
-    ("SCM_STRING_CHARS is deprecated.  See the manual for alternatives.");
-
-  /* We don't accept shared substrings here since they are not
-     null-terminated.
-  */
-  if (IS_SH_STRING (str))
-    scm_misc_error (NULL,
-		    "SCM_STRING_CHARS does not work with shared substrings",
-		    SCM_EOL);
-
-  /* We explicitly test for read-only strings to produce a better
-     error message.
-  */
-
-  if (IS_RO_STRING (str))
-    scm_misc_error (NULL,
-		    "SCM_STRING_CHARS does not work with read-only strings",
-		    SCM_EOL);
-
-  /* The following is still wrong, of course...
-   */
-  str = scm_i_string_start_writing (str);
-  chars = scm_i_string_writable_chars (str);
-  scm_i_string_stop_writing ();
-  return chars;
-}
-
-size_t
-scm_i_deprecated_string_length (SCM str)
-{
-  scm_c_issue_deprecation_warning
-    ("SCM_STRING_LENGTH is deprecated.  Use scm_c_string_length instead.");
-  return scm_c_string_length (str);
-}
-
-#endif
-
-static SCM
-string_handle_ref (scm_t_array_handle *h, size_t index)
-{
-  return scm_c_string_ref (h->array, index);
-}
-
-static void
-string_handle_set (scm_t_array_handle *h, size_t index, SCM val)
-{
-  scm_c_string_set_x (h->array, index, val);
-}
-
-static void
-string_get_handle (SCM v, scm_t_array_handle *h)
-{
-  h->array = v;
-  h->ndims = 1;
-  h->dims = &h->dim0;
-  h->dim0.lbnd = 0;
-  h->dim0.ubnd = scm_c_string_length (v) - 1;
-  h->dim0.inc = 1;
-  h->element_type = SCM_ARRAY_ELEMENT_TYPE_CHAR;
-  h->elements = h->writable_elements = NULL;
-}
-
-SCM_ARRAY_IMPLEMENTATION (scm_tc7_string, 0x7f,
-                          string_handle_ref, string_handle_set,
-                          string_get_handle)
 SCM_VECTOR_IMPLEMENTATION (SCM_ARRAY_ELEMENT_TYPE_CHAR, scm_make_string)
 
 void
